/* generate ec keypair */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdint.h>
#include <stdbool.h>

#include <openssl/err.h>
#include <openssl/bn.h>
#include <openssl/evp.h>
#include <openssl/x509.h>
#include <openssl/ec.h>

static void bufprintf(const char *label, const uint8_t *buf, size_t len)
{
	fprintf(stderr, "%s\n", label);
	for (int i = 0; i < len; i++)
		fprintf(stderr, "%02x ", buf[i] & 0xff);
	fprintf(stderr, "\n");
}

uint8_t *crypto_ec_key_get_subject_public_key(EVP_PKEY *pkey, int *outlen)
{
	unsigned char *der = NULL;
	EC_KEY *eckey;
	int der_len;
	uint8_t *out;

	*outlen = 0;
	eckey = EVP_PKEY_get1_EC_KEY(pkey);
	if (!eckey)
		return NULL;

	EC_KEY_set_conv_form(eckey, POINT_CONVERSION_COMPRESSED);
	der_len = i2d_PUBKEY(pkey, &der);
	EC_KEY_free(eckey);
	if (der_len <= 0) {
		fprintf(stderr, "openssl: i2d_PUBKEY() failed: %s\n",
			ERR_error_string(ERR_get_error(), NULL));
		return NULL;
	}

	out = calloc(1, der_len * sizeof(uint8_t));
	if (!out)
		return NULL;

	memcpy(out, der, der_len);
	*outlen = der_len;
	OPENSSL_free(der);

	return out;
}

uint8_t *crypto_ec_key_get_ecprivate_key(EVP_PKEY *pkey, bool include_pub,
					 int *outlen)
{
	unsigned char *der = NULL;
	unsigned int key_flags;
	EC_KEY *eckey;
	uint8_t *out;
	int der_len;

	*outlen = 0;
	eckey = EVP_PKEY_get1_EC_KEY(pkey);
	if (!eckey)
		return NULL;

	key_flags = EC_KEY_get_enc_flags(eckey);
	if (include_pub)
		key_flags &= ~EC_PKEY_NO_PUBKEY;
	else
		key_flags |= EC_PKEY_NO_PUBKEY;

	EC_KEY_set_enc_flags(eckey, key_flags);
	EC_KEY_set_conv_form(eckey, POINT_CONVERSION_UNCOMPRESSED);

	der_len = i2d_ECPrivateKey(eckey, &der);
	EC_KEY_free(eckey);
	if (der_len <= 0)
		return NULL;

	out = calloc(1, der_len * sizeof(uint8_t));
	if (!out)
		return NULL;

	memcpy(out, der, der_len);
	*outlen = der_len;
	OPENSSL_free(der);

	return out;
}

void crypto_ec_key_debug_print(const EVP_PKEY *key, const char *title)
{
	size_t rlen;
	BIO *out;
	char *txt;
	int res;

	out = BIO_new(BIO_s_mem());
	if (!out)
		return;

	EVP_PKEY_print_private(out, (EVP_PKEY *) key, 0, NULL);
	rlen = BIO_ctrl_pending(out);
	txt = calloc(1, (rlen + 1) * sizeof(char));
	if (txt) {
		res = BIO_read(out, txt, rlen);
		if (res > 0) {
			txt[res] = '\0';
			fprintf(stderr, "%s: %s\n", title, txt);
		}
		free(txt);
	}
	BIO_free(out);
}

void crypto_ec_key_print(const char *title, EVP_PKEY *key)
{
	uint8_t *der = NULL;
	int der_len = 0;

	crypto_ec_key_debug_print(key, title);

	der = crypto_ec_key_get_ecprivate_key(key, true, &der_len);
	if (der) {
		bufprintf("EC-PrivateKey", der, der_len);
	} else {
		der = crypto_ec_key_get_subject_public_key(key, &der_len);
		if (der)
			bufprintf("EC-PublicKey", der, der_len);
	}

	if (der)
		memset(der, 0, der_len);

	free(der);
}

EVP_PKEY *ec_keygen_group19(void)
{
	EVP_PKEY *params = NULL, *key = NULL;
	EC_KEY *ec_params = NULL, *eckey;
	EVP_PKEY_CTX *kctx = NULL;


	ec_params = EC_KEY_new_by_curve_name(NID_X9_62_prime256v1);
	if (!ec_params) {
		fprintf(stderr, "OpenSSL: Failed to generate EC_KEY parameters\n");
		goto fail;
	}

	EC_KEY_set_asn1_flag(ec_params, OPENSSL_EC_NAMED_CURVE);
	params = EVP_PKEY_new();
	if (!params || EVP_PKEY_set1_EC_KEY(params, ec_params) != 1) {
		fprintf(stderr, "OpenSSL: Failed to generate EVP_PKEY parameters\n");
		goto fail;
	}

	kctx = EVP_PKEY_CTX_new(params, NULL);
	if (!kctx || EVP_PKEY_keygen_init(kctx) != 1 || EVP_PKEY_keygen(kctx, &key) != 1) {
		fprintf(stderr, "OpenSSL: Failed to generate EC key\n");
		key = NULL;
		goto fail;
	}

	eckey = EVP_PKEY_get1_EC_KEY(key);
	if (!eckey) {
		key = NULL;
		goto fail;
	}

	EC_KEY_set_conv_form(eckey, POINT_CONVERSION_COMPRESSED);
	EC_KEY_free(eckey);

fail:
	EC_KEY_free(ec_params);
	EVP_PKEY_free(params);
	EVP_PKEY_CTX_free(kctx);

	return key;
}

int ec_gen_keypair(uint8_t **der, int *len)
{
	EVP_PKEY *key;

	key = ec_keygen_group19();
	if (!key)
		return -1;

	crypto_ec_key_print("Generated key", key);

	*der = crypto_ec_key_get_ecprivate_key(key, true, len);
	if (*der)
		bufprintf("EC-PrivateKey", *der, *len);

	EVP_PKEY_free(key);
	return 0;
}

#if 0
int main()
{
	EVP_PKEY *mykey;

	mykey = ec_gen_keypair();
	return 0;
}
#endif
